#include <config.h>

#include <sys/ioctl.h>

#include <directfb.h>

#include <direct/debug.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <core/state.h>
#include <core/surface.h>

#include <gfx/convert.h>

#include <mt53/mt53.h>

#include <gfx_if.h>

#include "mt53_2d.h"
#include "mt53_gfxdriver.h"


D_DEBUG_DOMAIN( MT53_2D, "MT53/2D", "MT53 2D Acceleration" );

/*
 * State validation flags.
 *
 * There's no prefix because of the macros below.
 */
enum {
     DESTINATION  = 0x00000001,
     COLOR        = 0x00000002,

     SOURCE       = 0x00000010,
     BLTOPTS_KEY  = 0x00000020,

     ALL          = 0x00000033
};

/*
 * State handling macros.
 */

#define MT53_VALIDATE(flags)        do { mdev->v_flags |=  (flags); } while (0)
#define MT53_INVALIDATE(flags)      do { mdev->v_flags &= ~(flags); } while (0)

#define MT53_CHECK_VALIDATE(flag)   do {                                           \
                                         if (! (mdev->v_flags & flag))             \
                                              mt53_validate_##flag( mdev, state );  \
                                    } while (0)

/**************************************************************************************************/

static bool mt53FillRectangle     ( void *drv, void *dev, DFBRectangle *rect );
static bool mt53FillRectangleBlend( void *drv, void *dev, DFBRectangle *rect );

static bool mt53Blit       ( void *drv, void *dev, DFBRectangle *srect, int dx, int dy );
static bool mt53BlitBlend  ( void *drv, void *dev, DFBRectangle *srect, int dx, int dy );

static bool mt53StretchBlit( void *drv, void *dev, DFBRectangle *srect, DFBRectangle *drect );

/**************************************************************************************************/

/*
 * Called by mt53SetState() to ensure that the destination registers are properly set
 * for execution of rendering functions.
 */
static inline void
mt53_validate_DESTINATION( MT53DeviceData *mdev,
                           CardState      *state )
{
     u32 color_mode = 0;

     /* Remember destination parameters for usage in rendering functions. */
     mdev->dst_addr   = state->dst.addr;
     mdev->dst_phys   = state->dst.phys;
     mdev->dst_pitch  = state->dst.pitch;
     mdev->dst_format = state->dst.buffer->format;
     mdev->dst_bpp    = DFB_BYTES_PER_PIXEL( mdev->dst_format );

     switch (mdev->dst_format) {
          case DSPF_RGB16:
               color_mode = CM_RGB565_DIRECT16;
               break;

          case DSPF_ARGB4444:
               color_mode = CM_ARGB4444_DIRECT16;
               break;

          case DSPF_ARGB:
               color_mode = CM_ARGB8888_DIRECT32;
               break;

          default:
               D_BUG( "unexpected destination format 0x%08x", mdev->dst_format );
     }

     GFX_SetDst( (UINT8*) mdev->dst_phys, color_mode, mdev->dst_pitch );

     /* Set the flag. */
     MT53_VALIDATE( DESTINATION );
}

/*
 * Called by mt53SetState() to ensure that the color register is properly set
 * for execution of rendering functions.
 */
static inline void
mt53_validate_COLOR( MT53DeviceData *mdev,
                     CardState      *state )
{
     switch (mdev->dst_format) {
          case DSPF_ARGB4444:
               mdev->color_pixel = PIXEL_ARGB4444( state->color.a,
                                                   state->color.r,
                                                   state->color.g,
                                                   state->color.b );
               break;

          case DSPF_ARGB:
               mdev->color_pixel = PIXEL_ARGB( state->color.a,
                                               state->color.r,
                                               state->color.g,
                                               state->color.b );
               break;

          default:
               D_BUG( "unexpected format %s", dfb_pixelformat_name(mdev->dst_format) );
     }

     GFX_SetColor( mdev->color_pixel );

     /* Set the flag. */
     MT53_VALIDATE( COLOR );
}

/*
 * Called by mt53SetState() to ensure that the source registers are properly set
 * for execution of blitting functions.
 */
static inline void
mt53_validate_SOURCE( MT53DeviceData *mdev,
                      CardState      *state )
{
     u32 color_mode = 0;

     /* Remember source parameters for usage in rendering functions. */
     mdev->src_addr   = state->src.addr;
     mdev->src_phys   = state->src.phys;
     mdev->src_pitch  = state->src.pitch;
     mdev->src_format = state->src.buffer->format;
     mdev->src_bpp    = DFB_BYTES_PER_PIXEL( mdev->src_format );

     switch (mdev->src_format) {
          case DSPF_RGB16:
               color_mode = CM_RGB565_DIRECT16;
               break;

          case DSPF_ARGB4444:
               color_mode = CM_ARGB4444_DIRECT16;
               break;

          case DSPF_ARGB:
               color_mode = CM_ARGB8888_DIRECT32;
               break;

          default:
               D_BUG( "unexpected source format 0x%08x", mdev->src_format );
     }

     GFX_SetSrc( (UINT8*) mdev->src_phys, color_mode, mdev->src_pitch );

     /* Set the flag. */
     MT53_VALIDATE( SOURCE );
}

/*
 * Called by mt53SetState() to ensure that the blitting options and color ke
 * are properly set for execution of blitting functions.
 */
static inline void
mt53_validate_BLTOPTS_KEY( MT53DeviceData *mdev,
                           CardState      *state )
{
     u32 bltopts = 0;
     u32 key_min = 0;
     u32 key_max = 0;

     if (mdev->src_format != mdev->dst_format)
          bltopts |= D_GFXFLAG_CFMT_ENA;

     if (state->blittingflags & DSBLIT_SRC_COLORKEY) {
          DFBColor key;

          bltopts |= D_GFXFLAG_TRANSPARENT;

          dfb_pixel_to_color( mdev->src_format, state->src_colorkey, &key );


          switch (mdev->src_format) {
               case DSPF_RGB16: /* ??? */
                    key_min = PIXEL_ARGB( 0x00, key.r & ~0x1f, key.g & ~0x3f, key.b & ~0x1f );
                    key_max = PIXEL_ARGB( 0xff, key.r |  0x1f, key.g |  0x3f, key.b |  0x1f );
                    break;

               case DSPF_ARGB4444: /* ??? */
                    key_min = PIXEL_ARGB( 0x00, key.r & ~0xf, key.g & ~0xf, key.b & ~0xf );
                    key_max = PIXEL_ARGB( 0xff, key.r |  0xf, key.g |  0xf, key.b |  0xf );
                    break;

               case DSPF_ARGB:
                    key_min = PIXEL_ARGB( 0x00, key.r, key.g, key.b );
                    key_max = PIXEL_ARGB( 0xff, key.r, key.g, key.b );
                    break;

               default:
                    D_BUG( "unexpected source format 0x%08x", mdev->src_format );
          }
     }

     GFX_SetBltOpt( bltopts, key_min, key_max );

     mdev->bltopts = bltopts;

     /* Set the flag. */
     MT53_VALIDATE( BLTOPTS_KEY );
}

/**************************************************************************************************/

/*
 * Wait for the blitter to be idle.
 *
 * This function is called before memory that has been written to by the hardware is about to be
 * accessed by the CPU (software driver) or another hardware entity like video encoder (by Flip()).
 * It can also be called by applications explicitly, e.g. at the end of a benchmark loop to include
 * execution time of queued commands in the measurement.
 */
DFBResult
mt53EngineSync( void *drv, void *dev )
{
     DFBResult       ret  = DFB_OK;
     MT53DriverData *mdrv = (MT53DriverData*) drv;

     if (ioctl( mdrv->fd, FBIO_GFX_WAIT ) < 0) {
          ret = errno2result( errno );

          D_PERROR( "MediaTek/Driver: FBIO_GFX_WAIT failed!\n" );
     }

     return ret;
}

/*
 * Reset the graphics engine.
 */
void
mt53EngineReset( void *drv, void *dev )
{
     GFX_SetAlpha( 0xff );
}

/*
 * Start processing of queued commands if required.
 *
 * This function is called before returning from the graphics core to the application.
 * Usually that's after each rendering function. The only functions causing multiple commands
 * to be queued with a single emition at the end are DrawString(), TileBlit(), BatchBlit(),
 * DrawLines() and possibly FillTriangle() which is emulated using multiple FillRectangle() calls.
 */
void
mt53EmitCommands( void *drv, void *dev )
{
     MT53DriverData *mdrv = (MT53DriverData*) drv;

     if (ioctl( mdrv->fd, FBIO_GFX_FLUSH ) < 0)
          D_PERROR( "MediaTek/Driver: FBIO_GFX_FLUSH failed!\n" );
}

/*
 * Check for acceleration of 'accel' using the given 'state'.
 */
void
mt53CheckState( void                *drv,
                void                *dev,
                CardState           *state,
                DFBAccelerationMask  accel )
{
     D_DEBUG_AT( MT53_2D, "mt53CheckState (state %p, accel 0x%08x) <- dest %p\n",
                 state, accel, state->destination );

     /* Return if the desired function is not supported at all. */
     if (accel & ~(MT53_SUPPORTED_DRAWINGFUNCTIONS | MT53_SUPPORTED_BLITTINGFUNCTIONS))
          return;

     /* Return if the destination format is not supported. */
     switch (state->destination->config.format) {
          case DSPF_RGB16:
          case DSPF_ARGB4444:
          case DSPF_ARGB:
               break;

          default:
               return;
     }

     /* Check if drawing or blitting is requested. */
     if (DFB_DRAWING_FUNCTION( accel )) {
          /* Return if unsupported drawing flags are set. */
          if (state->drawingflags & ~MT53_SUPPORTED_DRAWINGFLAGS)
               return;
     }
     else {
          /* Return if the source format is not supported. */
          switch (state->source->config.format) {
               case DSPF_RGB16:
               case DSPF_ARGB4444:
               case DSPF_ARGB:
                    /* FIXME: Conversion only without flags. */
                    if (!state->blittingflags ||
                        state->source->config.format == state->destination->config.format)
                         break;

               default:
                    return;
          }

          /* Return if unsupported blitting flags are set. */
          if (state->blittingflags & ~MT53_SUPPORTED_BLITTINGFLAGS)
               return;
     }

     /* Enable acceleration of the function. */
     state->accel |= accel;
}

/*
 * Make sure that the hardware is programmed for execution of 'accel' according to the 'state'.
 */
void
mt53SetState( void                *drv,
              void                *dev,
              GraphicsDeviceFuncs *funcs,
              CardState           *state,
              DFBAccelerationMask  accel )
{
     MT53DeviceData         *mdev     = (MT53DeviceData*) dev;
     StateModificationFlags  modified = state->mod_hw;

     D_DEBUG_AT( MT53_2D, "mt53SetState (state %p, accel 0x%08x) <- dest %p, modified 0x%08x\n",
                 state, accel, state->destination, modified );

     /*
      * 1) Invalidate hardware states
      *
      * Each modification to the hw independent state invalidates one or more hardware states.
      */

     /* Simply invalidate all? */
     if (modified == SMF_ALL) {
          MT53_INVALIDATE( ALL );
     }
     else if (modified) {
          /* Invalidate destination registers. */
          if (modified & SMF_DESTINATION)
               MT53_INVALIDATE( DESTINATION | COLOR );
          else if (modified & SMF_COLOR)
               MT53_INVALIDATE( COLOR );

          /* Invalidate source registers. */
          if (modified & SMF_SOURCE)
               MT53_INVALIDATE( SOURCE | BLTOPTS_KEY );
          else if (modified & (SMF_SRC_COLORKEY | SMF_BLITTING_FLAGS))
               MT53_INVALIDATE( BLTOPTS_KEY );
     }

     /*
      * 2) Validate hardware states
      *
      * Each function has its own set of states that need to be validated.
      */

     /* Always requiring valid destination... */
     MT53_CHECK_VALIDATE( DESTINATION );

     /* Remember color. */
     mdev->color = state->color;

     /* Depending on the function... */
     switch (accel) {
          case DFXL_FILLRECTANGLE:
               /* ...require valid drawing color. */
               MT53_CHECK_VALIDATE( COLOR );

               if (state->drawingflags & DSDRAW_BLEND)
                    funcs->FillRectangle = mt53FillRectangleBlend;
               else
                    funcs->FillRectangle = mt53FillRectangle;

               if (mdev->bltopts) {
                    mdev->bltopts = 0;

                    GFX_SetBltOpt( 0, 0, 0 );

                    MT53_INVALIDATE( BLTOPTS_KEY );
               }

               /*
                * 3) Tell which functions can be called without further validation, i.e. SetState()
                *
                * When the hw independent state is changed, this collection is reset.
                */
               state->set = MT53_SUPPORTED_DRAWINGFUNCTIONS;
               break;

          case DFXL_BLIT:
          case DFXL_STRETCHBLIT:
               /* ...require valid source and blitting options. */
               MT53_CHECK_VALIDATE( SOURCE );
               MT53_CHECK_VALIDATE( BLTOPTS_KEY );

               /* Use alpha blending? */
               if (state->blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)) {
                    if (!(state->blittingflags & DSBLIT_BLEND_COLORALPHA))
                         mdev->color.a = 0xff;

                    funcs->Blit = mt53BlitBlend;
               }
               else
                    funcs->Blit = mt53Blit;

               funcs->StretchBlit = mt53StretchBlit;

               /*
                * 3) Tell which functions can be called without further validation, i.e. SetState()
                *
                * When the hw independent state is changed, this collection is reset.
                */
               state->set = MT53_SUPPORTED_BLITTINGFUNCTIONS;
               break;

          default:
               D_BUG( "unexpected drawing/blitting function" );
               break;
     }

     /*
      * 4) Clear modification flags
      *
      * All flags have been evaluated in 1) and remembered for further validation.
      * If the hw independent state is not modified, this function won't get called
      * for subsequent rendering functions, unless they aren't defined by 3).
      */
     state->mod_hw = 0;
}

/*
 * Render a filled rectangle using the current hardware state.
 */
static bool
mt53FillRectangle( void *drv, void *dev, DFBRectangle *rect )
{
     D_DEBUG_AT( MT53_2D, "%s( %d,%d-%dx%d )\n", __FUNCTION__, DFB_RECTANGLE_VALS( rect ) );

     GFX_Fill( DFB_RECTANGLE_VALS( rect ) );

     return true;
}

static bool
mt53FillRectangleBlend( void *drv, void *dev, DFBRectangle *rect )
{
     MT53DeviceData *mdev = (MT53DeviceData*) dev;

     D_DEBUG_AT( MT53_2D, "%s( %d,%d-%dx%d )\n", __FUNCTION__, DFB_RECTANGLE_VALS( rect ) );

     GFX_ComposeLoop( 0, 0, DFB_RECTANGLE_VALS( rect ), 
                      mdev->color.a, ROP_SRC_OVER, 1 );

     return true;
}

/*
 * Copy a rectangle using the current hardware state.
 */
static bool
mt53Blit( void *drv, void *dev, DFBRectangle *srect, int dx, int dy )
{
     D_DEBUG_AT( MT53_2D, "%s( %d,%d-%dx%d -> %d, %d )\n", __FUNCTION__,
                 DFB_RECTANGLE_VALS( srect ), dx, dy );

     GFX_BitBlt( srect->x, srect->y, dx, dy, srect->w, srect->h );

     return true;
}

static bool
mt53BlitBlend( void *drv, void *dev, DFBRectangle *srect, int dx, int dy )
{
     MT53DeviceData *mdev = (MT53DeviceData*) dev;

     D_DEBUG_AT( MT53_2D, "%s( %d,%d-%dx%d -> %d, %d )\n", __FUNCTION__,
                 DFB_RECTANGLE_VALS( srect ), dx, dy );

     GFX_ComposeLoop( srect->x, srect->y, dx, dy, srect->w, srect->h, 
                      mdev->color.a, ROP_SRC_OVER, 0 );

     return true;
}

/*
 * Stretch a rectangle using the current hardware state.
 */
static bool
mt53StretchBlit( void *drv, void *dev, DFBRectangle *srect, DFBRectangle *drect )
{
     D_DEBUG_AT( MT53_2D, "%s( %d,%d-%dx%d -> %d, %d )\n", __FUNCTION__,
                 DFB_RECTANGLE_VALS( srect ), dx, dy );

     GFX_StretchBlt( DFB_RECTANGLE_VALS( srect ), DFB_RECTANGLE_VALS( drect ) );

     return true;
}

